package iTunesBackup;

import org.w3c.dom.*;
import javax.swing.*;

import java.io.*;

import javax.xml.parsers.*;
import javax.xml.xpath.*;
import java.util.*;
import java.net.*;
import org.xml.sax.InputSource;


/**
 * Tune Backup
 * (c) 2007 Gabe Albert
 * Gabe Albert - gabeosx (at) yahoo.com - www.gabealbert.info
 * The use and distribution of this work is subject to the terms of the 
 * GNU GPL license.  A full-text copy of the license is available online at
 * http://www.gnu.org/licenses/gpl.txt
 * @author Gabe Albert
 */
public class BackerUpper extends Thread implements Runnable {
	public static TrackCollection theTracks = new TrackCollection();
	public static final int DISPLAY_TRACK = 1;
	public static final int COPY_TRACK = 2;
	public static final int MANAGE_THREADS = 3;
	private String destinationBaseDir;
	private int theTask=0;
	private MusicTrack keyTrack = new MusicTrack();
	public static TuneController tc = null;
	private String iTunesXMLPath = new String();
	int sCount;
	DBPipe dbPipe;
	private String currentSong = "";
	
	private WeakHashMap<Integer, Thread>threads;

	public BackerUpper(String xmlPath) {
		/*since this is static, we will check if our tracks are populated or populate them, since it
    	is necessary in order to do anything*/
		this.iTunesXMLPath=xmlPath;
		if(theTracks.size()==0)
			exclusiveSetItOff();
	}

	public BackerUpper(TuneController tc, String xmlPath) {
		this.tc=tc;
		this.iTunesXMLPath=xmlPath;
		if(theTracks.size()==0)
			exclusiveSetItOff();
	}

	/**
	 * This constructor is used for threads, so we know which
	 * function to execute
	 * @param whatFunction
	 * @param xmlPath TODO
	 */
	public BackerUpper(int whatFunction, String xmlPath) {
		this.theTask=whatFunction;
		this.iTunesXMLPath = xmlPath;
		if(theTracks.size()==0)
			exclusiveSetItOff();
	}

	/**
	 * Another constructor for threads that require a key
	 * @param whatFunction
	 * @param xmlPath TODO
	 * @param theKey
	 */
	public BackerUpper(int whatFunction, MusicTrack keyTrack, String xmlPath, String destinationDir, DBPipe dbPipe) {
		this.theTask=whatFunction;
		this.keyTrack=keyTrack;
		this.destinationBaseDir=destinationDir;
		this.iTunesXMLPath=xmlPath;
		this.dbPipe=dbPipe;
		if(theTracks.size()==0)
			exclusiveSetItOff();
	}

	public BackerUpper(int whatFunction, String xmlPath, String baseDir) {
		this.theTask=whatFunction;
		this.keyTrack=keyTrack;
		this.destinationBaseDir=baseDir;
		this.iTunesXMLPath=xmlPath;
		if(theTracks.size()==0)
			exclusiveSetItOff();
	}
	
	/**
	 * Parses XML
	 * This must be done exclusively, in order to avoid db concurrency issues.
	 *
	 */
	public void exclusiveSetItOff() {
		if(this.getClass().toString().equalsIgnoreCase(""))
		try {
			SwingUtilities.invokeAndWait(new Runnable() {
				public void run() {
					tc.setProgressString("Analyzing library, please wait.");
				}
			});
		}
		catch(Exception ex) {
			ex.printStackTrace();
		}
		try {
			SwingUtilities.invokeAndWait(new Runnable() {
				public void run() {
					if(theTracks.size()==0) setItOff();
				}
			});
		}
		catch(Exception ex) {
			ex.printStackTrace();
		}
	}

	public BackerUpper(int whatFunction, String xmlPath, String baseDir, TuneController theTC) {
		this.theTask=whatFunction;
		this.keyTrack=keyTrack;
		this.destinationBaseDir=baseDir;
		this.iTunesXMLPath=xmlPath;
		this.tc=theTC;
	}

	public void run() {
		if(theTracks.size()==0)
			exclusiveSetItOff();
		if(theTask==COPY_TRACK) {
			copyTrack(keyTrack, destinationBaseDir);
		}
		else if(theTask==DISPLAY_TRACK) {
			printTrack(keyTrack);
		}
		else if(theTask==MANAGE_THREADS) {
			this.spawnThreads();
		}
	}

	public void spawnThreads() {
		dbPipe = new DBPipe();
		sCount=0;
		System.out.println("theTracks.dirtyTracks.size()"+theTracks.dirtyTracks.size());
		tc.setProgressMax(theTracks.dirtyTracks.size());
		MusicTrack theTrack;
		int max = theTracks.dirtyTracks.size();
		while(theTracks.dirtyTracks.size()>0) {
			theTrack=theTracks.getByTrackId(Integer.parseInt(theTracks.dirtyTracks.pop()));
			if(theTrack!=null)
				copyTrack(theTrack, destinationBaseDir);

			sCount=max-theTracks.dirtyTracks.size();
			currentSong = theTrack.getTrackArtist()+" - "+theTrack.getTrackName();
			updateProgressBar();
		}
		if(theTracks.dirtyTracks.size()==0) {
			updateProgressBar();
		}
	}

	public void updateProgressBar() {
		Runnable setProgressBar = new Runnable() {
			public void run() {
				tc.setProgressCurrent(sCount);
				tc.setProgressString(currentSong);
			}
		};
		try {
			SwingUtilities.invokeAndWait(setProgressBar);
		}
		catch(Exception e) {

		}
	}

	public int getFirstDeadThread() {
		int thread=-1;
		while(thread==-1) {
			for(int i=0; i<threads.size(); i++) {
				if(threads.get(i)!=null && !threads.get(i).isAlive()) {
					thread=i;
					try {
						threads.get(i).join();
					}
					catch(Exception ex) {
						ex.printStackTrace();
					}
					return i;
				}
				else if(threads.get(i)==null) {
					thread=i;
					return i;
				}
			}
			if(threads.size()==0) {
				thread=0;
				return 0;
			}
		}
		return thread;
	}

	public void printTrack(MusicTrack trackToPrint) {
		if(dbPipe.performAction(trackToPrint)==DBPipe.track_isnew) {
			System.out.println(trackToPrint.getTrackArtist()+" - "+trackToPrint.getTrackName()+"... "+trackToPrint.getTrackTotalTime()+"ms "+trackToPrint.getTrackPlayCount()+"x");
			dbPipe.addTrack(trackToPrint);
		}
	}

	public void setBaseDir(String baseDir) {
		this.destinationBaseDir=baseDir;
	}

	public void copyTrack(MusicTrack theTrack, String destinationDirBase) {
		try {//.replace(" ", "%20")
			//URL destinationURL = new URL("file:///"+destinationDirBase.replace(" ", "%20")+"/"+theTrack.getSafeFilePath());
			//theTrack.setTrackBackupLocation(new URI(destinationURL.toString()).toString());
			theTrack.setTrackBackupLocation(new URI("file://"+destinationBaseDir.replace(" ", "%20")+"/"+theTrack.getSafeFilePath().toString()));
		}
		catch(Exception ex) {
			ex.printStackTrace();
		}
		if(dbPipe.performAction(theTrack)==DBPipe.track_isnew) {
			if(theTrack.getTrackKind()!=null && !theTrack.getTrackKind().contains("MPEG audio stream")) { //MPEG audio stream is a streaming audio server.  We don't want to copy this.
				//System.out.println("Copying "+theTrack.getTrackLocation().toString()+" to "+(theTrack.getTrackBackupLocation()));
				try {
					File in = new File(new URI(theTrack.getTrackLocation().replace("localhost", "")));
					File out = new File(theTrack.getTrackBackupLocation());
					File dir = new File(out.getAbsolutePath().substring(0, out.getAbsolutePath().lastIndexOf("/")+1));
					dir.mkdirs();
					out.createNewFile();
					BufferedInputStream bis = new BufferedInputStream(new FileInputStream(in));
					BufferedOutputStream bos = new BufferedOutputStream(new FileOutputStream(out));
					byte inArr[] = new byte[bis.available()];
					int size = bis.read(inArr,0,inArr.length);
					System.out.println("Size is :"+size);
					bos.write(inArr,0,size);
					bis.close();
					bos.close();
					dbPipe.addTrack(theTrack); // update the table
				} catch(Exception e) {
					e.printStackTrace();
					theTrack.setErrorMessage(e.getMessage());
					tc.getBadTracks().put(theTrack);
				}
			}
		}
	}

	public void setKey(MusicTrack theTrack) {
		this.keyTrack = theTrack;
	}

	public void setItOff() {
		Date startDate = new Date();
		populateWithSAX();
		try {
			SwingUtilities.invokeAndWait(new Runnable() {
				public void run() {
					tc.setProgressString("Syncing... this may take a while");
				}
			});
		}
		catch(Exception ex) {
			ex.printStackTrace();
		}
		Date endDate = new Date();
		long totalTime=0;
		for(int i=0; i<theTracks.size(); i++) {
			totalTime+=((Integer.parseInt(theTracks.get(i).getTrackTotalTime())/60000)*Integer.parseInt(theTracks.get(i).getTrackPlayCount()));
		}
		System.out.println("Aren't you jealous?  I have "+theTracks.size()+" tracks in my iTunes library.");
		System.out.println("XML traversed in "+((endDate.getTime()-startDate.getTime())/1000)+" seconds");
		System.out.println("I've spent "+totalTime+" minutes listening to music.");
	}

	/**
	 *
	 * @deprecated
	 */
	public void getTrackInfo() {
		String theExpr = "/plist/dict/key[text()='Tracks']/following-sibling::dict/dict";
		try {
			DocumentBuilderFactory domFactory = DocumentBuilderFactory.newInstance();
			domFactory.setNamespaceAware(true);
			DocumentBuilder builder = domFactory.newDocumentBuilder();
			if(tc!=null)
				tc.setProgressString("Loading XML file");
			Document doc = builder.parse("/Users/gabe/Music/iTunes/iTunes Music Library.xml");
			if(tc!=null)
				tc.setProgressString("Traversing XML");
			XPathFactory factory = XPathFactory.newInstance();
			XPath xpath = factory.newXPath();
			XPathExpression expr = xpath.compile(theExpr);
			Object result = expr.evaluate(doc, XPathConstants.NODESET);
			NodeList nodes = (NodeList) result;
			if(tc!=null)
				tc.setProgressString("Reading track info");
			for (int trackNodeCt = 0; trackNodeCt < nodes.getLength(); trackNodeCt++) {
				MusicTrack tempTrack = new MusicTrack();
				if(tc!=null) {
					tc.setProgressCurrent(0);
					tc.setProgressString("Getting track "+trackNodeCt+"/"+nodes.getLength());
					tc.setProgressMax(nodes.getLength());

					tc.setProgressCurrent(trackNodeCt);
				}
				for(int trackElement=0; trackElement<nodes.item(trackNodeCt).getChildNodes().getLength(); trackElement++) {
					Node currentNode = nodes.item(trackNodeCt).getChildNodes().item(trackElement);
					if(currentNode.getNodeName().equalsIgnoreCase("key")) {
						if(currentNode.getTextContent().equals("Track ID"))
							tempTrack.setTrackId(currentNode.getNextSibling().getTextContent());
						if(currentNode.getTextContent().equalsIgnoreCase("Name"))
							tempTrack.setTrackName(currentNode.getNextSibling().getTextContent());
						if(currentNode.getTextContent().equalsIgnoreCase("Artist"))
							tempTrack.setTrackArtist(currentNode.getNextSibling().getTextContent());
						if(currentNode.getTextContent().equalsIgnoreCase("Album"))
							tempTrack.setTrackAlbum(currentNode.getNextSibling().getTextContent());
						if(currentNode.getTextContent().equalsIgnoreCase("Composer"))
							tempTrack.setTrackComposer(currentNode.getNextSibling().getTextContent());
						if(currentNode.getTextContent().equalsIgnoreCase("Genre"))
							tempTrack.setTrackGenre(currentNode.getNextSibling().getTextContent());
						if(currentNode.getTextContent().equalsIgnoreCase("Kind"))
							tempTrack.setTrackKind(currentNode.getNextSibling().getTextContent());
						if(currentNode.getTextContent().equalsIgnoreCase("Size"))
							tempTrack.setTrackSize(currentNode.getNextSibling().getTextContent());
						if(currentNode.getTextContent().equalsIgnoreCase("Total Time"))
							tempTrack.setTrackTotalTime(currentNode.getNextSibling().getTextContent());
						if(currentNode.getTextContent().equalsIgnoreCase("Disc Number"))
							tempTrack.setTrackDiscNumber(currentNode.getNextSibling().getTextContent());
						if(currentNode.getTextContent().equalsIgnoreCase("Disc Count"))
							tempTrack.setTrackDiscCount(currentNode.getNextSibling().getTextContent());
						if(currentNode.getTextContent().equalsIgnoreCase("Year"))
							tempTrack.setTrackYear(currentNode.getNextSibling().getTextContent());
						if(currentNode.getTextContent().equalsIgnoreCase("Date Modified"))
							tempTrack.setTrackDateModified(currentNode.getNextSibling().getTextContent());
						if(currentNode.getTextContent().equalsIgnoreCase("Date Added"))
							tempTrack.setTrackDateAdded(currentNode.getNextSibling().getTextContent());
						if(currentNode.getTextContent().equalsIgnoreCase("Bit Rate"))
							tempTrack.setTrackBitRate(currentNode.getNextSibling().getTextContent());
						if(currentNode.getTextContent().equalsIgnoreCase("Sample Rate"))
							tempTrack.setTrackSampleRate(currentNode.getNextSibling().getTextContent());
						if(currentNode.getTextContent().equalsIgnoreCase("Persistent ID"))
							tempTrack.setTrackPersistentId(currentNode.getNextSibling().getTextContent());
						if(currentNode.getTextContent().equalsIgnoreCase("Track Type"))
							tempTrack.setTrackTrackType(currentNode.getNextSibling().getTextContent());
						if(currentNode.getTextContent().equalsIgnoreCase("File Type"))
							tempTrack.setTrackFileType(currentNode.getNextSibling().getTextContent());
						if(currentNode.getTextContent().equalsIgnoreCase("File Creator"))
							tempTrack.setTrackFileCreator(currentNode.getNextSibling().getTextContent());
						if(currentNode.getTextContent().equalsIgnoreCase("Location"))
							tempTrack.setTrackLocation(currentNode.getNextSibling().getTextContent());
						if(currentNode.getTextContent().equalsIgnoreCase("File Folder Count"))
							tempTrack.setTrackFileFolderCount(currentNode.getNextSibling().getTextContent());
						if(currentNode.getTextContent().equalsIgnoreCase("Library Folder Count"))
							tempTrack.setTrackLibraryFolderCount(currentNode.getNextSibling().getTextContent());
						if(currentNode.getTextContent().equalsIgnoreCase("Play Count"))
							tempTrack.setTrackPlayCount(currentNode.getNextSibling().getTextContent());
						if(currentNode.getTextContent().equalsIgnoreCase("Rating"))
							tempTrack.setTrackRating(currentNode.getNextSibling().getTextContent());
					}
				}
				theTracks.put(tempTrack);
			}
		}
		catch(Exception e) {
			e.printStackTrace();
		}
	}

	public HashMap<String, PropertyHandler> handlerCollection = new HashMap();

	public void populateElementCollection() {
		handlerCollection.put("Track ID",
				new PropertyHandler() {
			public void execute(MusicTrack tempTrack, String theKey) {
				tempTrack.setTrackId(theKey);
			}
		}
		);
		handlerCollection.put("Name",
				new PropertyHandler() {
			public void execute(MusicTrack tempTrack, String theKey) {
				tempTrack.setTrackName(theKey);
			}
		}
		);
		handlerCollection.put("Artist",
				new PropertyHandler() {
			public void execute(MusicTrack tempTrack, String theKey) {
				tempTrack.setTrackArtist(theKey);
			}
		}
		);
		handlerCollection.put("Album",
				new PropertyHandler() {
			public void execute(MusicTrack tempTrack, String theKey) {
				tempTrack.setTrackAlbum(theKey);
			}
		}
		);
		handlerCollection.put("Composer",
				new PropertyHandler() {
			public void execute(MusicTrack tempTrack, String theKey) {
				tempTrack.setTrackComposer(theKey);
			}
		}
		);
		handlerCollection.put("Genre",
				new PropertyHandler() {
			public void execute(MusicTrack tempTrack, String theKey) {
				tempTrack.setTrackGenre(theKey);
			}
		}
		);
		handlerCollection.put("Kind",
				new PropertyHandler() {
			public void execute(MusicTrack tempTrack, String theKey) {
				tempTrack.setTrackGenre(theKey);
			}
		}
		);
		handlerCollection.put("Size",
				new PropertyHandler() {
			public void execute(MusicTrack tempTrack, String theKey) {
				tempTrack.setTrackSize(theKey);
			}
		}
		);
		handlerCollection.put("Total Time",
				new PropertyHandler() {
			public void execute(MusicTrack tempTrack, String theKey) {
				tempTrack.setTrackTotalTime(theKey);
			}
		}
		);
		handlerCollection.put("Disc Number",
				new PropertyHandler() {
			public void execute(MusicTrack tempTrack, String theKey) {
				tempTrack.setTrackDiscNumber(theKey);
			}
		}
		);
		handlerCollection.put("Disc Count",
				new PropertyHandler() {
			public void execute(MusicTrack tempTrack, String theKey) {
				tempTrack.setTrackDiscCount(theKey);
			}
		}
		);
		handlerCollection.put("Year",
				new PropertyHandler() {
			public void execute(MusicTrack tempTrack, String theKey) {
				tempTrack.setTrackYear(theKey);
			}
		}
		);
		handlerCollection.put("Date Modified",
				new PropertyHandler() {
			public void execute(MusicTrack tempTrack, String theKey) {
				tempTrack.setTrackDateModified(theKey);
			}
		}
		);
		handlerCollection.put("Date Added",
				new PropertyHandler() {
			public void execute(MusicTrack tempTrack, String theKey) {
				tempTrack.setTrackDateAdded(theKey);
			}
		}
		);
		handlerCollection.put("Bit Rate",
				new PropertyHandler() {
			public void execute(MusicTrack tempTrack, String theKey) {
				tempTrack.setTrackBitRate(theKey);
			}
		}
		);
		handlerCollection.put("Sample Rate",
				new PropertyHandler() {
			public void execute(MusicTrack tempTrack, String theKey) {
				tempTrack.setTrackSampleRate(theKey);
			}
		}
		);
		handlerCollection.put("Persistent ID",
				new PropertyHandler() {
			public void execute(MusicTrack tempTrack, String theKey) {
				tempTrack.setTrackPersistentId(theKey);
			}
		}
		);
		handlerCollection.put("Track Type",
				new PropertyHandler() {
			public void execute(MusicTrack tempTrack, String theKey) {
				tempTrack.setTrackTrackType(theKey);
			}
		}
		);
		handlerCollection.put("File Type",
				new PropertyHandler() {
			public void execute(MusicTrack tempTrack, String theKey) {
				tempTrack.setTrackFileType(theKey);
			}
		}
		);
		handlerCollection.put("File Creator",
				new PropertyHandler() {
			public void execute(MusicTrack tempTrack, String theKey) {
				tempTrack.setTrackFileCreator(theKey);
			}
		}
		);
		handlerCollection.put("Location",
				new PropertyHandler() {
			public void execute(MusicTrack tempTrack, String theKey) {
				tempTrack.setTrackLocation(theKey);
			}
		}
		);
		handlerCollection.put("File Folder Count",
				new PropertyHandler() {
			public void execute(MusicTrack tempTrack, String theKey) {
				tempTrack.setTrackFileFolderCount(theKey);
			}
		}
		);
		handlerCollection.put("Library Folder Count",
				new PropertyHandler() {
			public void execute(MusicTrack tempTrack, String theKey) {
				tempTrack.setTrackLibraryFolderCount(theKey);
			}
		}
		);
		handlerCollection.put("Play Count",
				new PropertyHandler() {
			public void execute(MusicTrack tempTrack, String theKey) {
				tempTrack.setTrackPlayCount(theKey);
			}
		}
		);
		handlerCollection.put("Rating",
				new PropertyHandler() {
			public void execute(MusicTrack tempTrack, String theKey) {
				tempTrack.setTrackRating(theKey);
			}
		}
		);
		handlerCollection.put("Kind", new PropertyHandler() {
			public void execute(MusicTrack tempTrack, String theKey) {
				tempTrack.setTrackKind(theKey);
			}
		});

	}

	public void populateWithSAX() {
		this.populateElementCollection();
		System.out.println("trying with: "+iTunesXMLPath);
		try {
			PropertyListHandlerImpl plh = new PropertyListHandlerImpl(theTracks, handlerCollection);
			PropertyListParser plist = new PropertyListParser(plh, null);
			plist.parse(new InputSource(new FileInputStream(iTunesXMLPath)));
		}
		catch(Exception ex) {
			ex.printStackTrace();
		}
	}

}
