package iTunesBackup;

import java.sql.*;
import java.util.HashMap;

import org.h2.*;
/**
 * Tune Backup
 * (c) 2007 Gabe Albert
 * Gabe Albert - gabeosx (at) yahoo.com - www.gabealbert.info
 * The use and distribution of this work is subject to the terms of the 
 * GNU GPL license.  A full-text copy of the license is available online at
 * http://www.gnu.org/licenses/gpl.txt
 * @author Gabe Albert
 */
public class DBPipe {

	private Connection conn=null;

	public static final int track_unchanged = 1;
	public static final int track_updated = 2;
	public static final int track_isnew = 3;

	private Statement sqlStatement = null;

	private static final boolean DEBUG = false;

	private static final String trackInfoQuery = "SELECT ID, UPDATED FROM TRACKS WHERE ID = ";
	
	private HashMap<String, Integer>cachedTrackAction = new HashMap<String, Integer>();

	public DBPipe() {
		startConnect();
	}

	public void startConnect() {
		try {
			if(conn==null || conn.isClosed()) {
				Class.forName("org.h2.Driver");
				conn = DriverManager.getConnection("jdbc:h2:test", "sa", "");
				sqlStatement = conn.createStatement();
				if(DEBUG) System.out.println("connected to database");
				try {
					conn.prepareStatement("SELECT * FROM tracks").execute();
				}
				catch(Exception ex) {
					System.err.println("No tracks table exists, creating one");
					try {
						conn.prepareStatement("create table tracks(ID VARCHAR2(50),UPDATED VARCHAR2(1000),OLDDIR VARCHAR2(1000),NEWDIR VARCHAR2(1000));").execute();
					}
					catch(Exception createex) {
						createex.printStackTrace();
					}
				}
				try {
					conn.prepareStatement("select * from prefs").execute();
				}
				catch(Exception ex) {
					System.err.println("No prefs table exists, creating one");
					try {
						conn.prepareStatement("create table prefs(PREFNAME VARCHAR2(50), PREFVALUE VARCHAR2(1000));").execute();
					}
					catch(Exception createex) {
						createex.printStackTrace();
					}
				}
			}
		} catch(Exception ex) {
			ex.printStackTrace();
		}
	}

	public ResultSet getResultsByQuery(String queryString) {
		startConnect();
		ResultSet rs=null;
		try {
			rs = sqlStatement.executeQuery(queryString);
			if(DEBUG) System.out.println("executed query: "+queryString);
		}
		catch(Exception ex) {
			ex.printStackTrace();
		}
		return rs;
	}

	public void debugResults(ResultSet rs) {
		try {
			while(rs.next()) {
				System.out.println(rs.getString(1)+", "+rs.getString(2));
			}
			if(DEBUG) System.out.println("end debugResults()");
		}
		catch(Exception ex) {
			ex.printStackTrace();
		}
	}

	public void testDBPipe() {
		startConnect();
		System.out.println("----SPILLING CONTENTS OF DB----");
		debugResults(getResultsByQuery("SELECT ID, UPDATED FROM TRACKS"));
		MusicTrack testTrack = new MusicTrack();
		try {
			conn.prepareStatement("DELETE FROM TRACKS").execute();
			System.out.println("----DATABASE CLEARED----");
		}
		catch(Exception ex) {

		}
		testTrack.setTrackId("1");
		testTrack.setTrackSize("1");
		System.out.println("----TESTING MANUAL ADD----");
		addTrack(testTrack);
		// make sure update works
		System.out.println("----TESTING AUTOMATIC UPDATE----");
		testTrack.setTrackSize("2");
		performAction(testTrack);
		System.out.println("----TESTING AUTOMATIC ADD----");
		testTrack.setTrackId("2");
		performAction(testTrack);
		System.out.println("----TESTING NO CHANGE ACTION----");
		performAction(testTrack);
		System.out.println("----END TESTS----");
	}

	public int checkTrack(MusicTrack theTrack) {
		startConnect();
		String checkQuery = trackInfoQuery+theTrack.getTrackId();
		String trackSize="";
		if(cachedTrackAction.get(theTrack.getTrackId())!=null)
			return cachedTrackAction.get(theTrack.getTrackId());
		else {
			try {
				ResultSet results = getResultsByQuery(checkQuery);

				while(results.next()) {
					if(trackSize.equalsIgnoreCase(theTrack.getTrackSize())) {
						if(DEBUG) System.out.println("result: "+track_updated+" from "+checkQuery);
						cachedTrackAction.put(theTrack.getTrackId(), track_updated);
						return track_updated;
					}
					else {
						if(DEBUG) System.out.println("result: "+track_unchanged+" from "+checkQuery);
						cachedTrackAction.put(theTrack.getTrackId(), track_unchanged);
						return track_unchanged;
					}
				}
			}
			catch(Exception ex) {
				//ex.printStackTrace();
				System.err.println("ERROR ON: "+checkQuery);
			}
		}
		
		cachedTrackAction.put(theTrack.getTrackId(), track_isnew);
		if(DEBUG) System.out.println("result: "+track_isnew+" from "+checkQuery);
		return track_isnew;
	}

	public void addTrack(MusicTrack theTrack) {
		startConnect();
		PreparedStatement addStmt;
		cachedTrackAction.put(theTrack.getTrackId(), track_unchanged);
		try {
			// ID VARCHAR2(50),UPDATED VARCHAR2(1000),OLDDIR VARCHAR2(1000),NEWDIR VARCHAR2(1000)
			addStmt = conn.prepareStatement("INSERT INTO TRACKS VALUES (?, ?, ?, ?)");
			addStmt.setString(1, theTrack.getTrackId());
			addStmt.setString(2, theTrack.getTrackSize());
			addStmt.setString(3, theTrack.getTrackLocation());
			addStmt.setString(4, theTrack.getTrackBackupLocation().toString());
			addStmt.execute();
		}
		catch(Exception ex) {

		}
	}

	public void updateTrack(MusicTrack theTrack) {
		startConnect();
		cachedTrackAction.put(theTrack.getTrackId(), track_unchanged);
		PreparedStatement updtStmt;
		try {
			updtStmt = conn.prepareStatement("UPDATE TRACKS SET UPDATED=?, OLDDIR=?, NEWDIR=? WHERE ID=?");
			updtStmt.setString(1, theTrack.getTrackSize());
			updtStmt.setString(2, theTrack.getTrackLocation());
			updtStmt.setString(3, theTrack.getTrackBackupLocation().toString());
			updtStmt.setString(4, theTrack.getTrackId());
			updtStmt.execute();
		}
		catch(Exception ex) {
			ex.printStackTrace();
		}
	}

	public int performAction(MusicTrack theTrack) {
		startConnect();
		if(checkTrack(theTrack)==track_unchanged) {
			return track_unchanged;
		}
		else if(checkTrack(theTrack)==track_updated) {
			updateTrack(theTrack);
			return track_updated;
		}
		else if(checkTrack(theTrack)==track_isnew) {
			return track_isnew;
		}
		if(DEBUG) System.out.println("no action necessary - no changes");
		return track_unchanged;
	}

	public static void main(String[] args) {
		DBPipe dbp = new DBPipe();
		dbp.testDBPipe();
	}

	public void setPref(String prefName, String prefValue) {
		startConnect();
		String currValue = getPref(prefName);
		String theQuery = "";
		if(!currValue.equalsIgnoreCase("") && currValue!=null) {
			theQuery = "UPDATE PREFS SET PREFVALUE = '"+prefValue+"' where PREFNAME = '"+prefName+"'";
		}
		else {
			theQuery = "INSERT INTO PREFS VALUES('"+prefName+"', '"+prefValue+"')";
		}
		try {
			conn.prepareStatement(theQuery).execute();
		}
		catch(Exception ex) {
			ex.printStackTrace();
		}
		if(DEBUG) System.out.println("setPref(): set "+prefName+" to "+prefValue+" -- query used: "+theQuery);
	}

	public String getPref(String prefName) {
		startConnect();
		ResultSet rs = getResultsByQuery("SELECT PREFVALUE FROM PREFS WHERE PREFNAME = '"+prefName+"'");
		String prefValue = "";
		try {
			while(rs.next())
				prefValue = rs.getString(1);
		}
		catch(Exception ex) {
			ex.printStackTrace();
		}
		if(DEBUG) System.out.println("getPref(): "+prefName+"="+prefValue);
		return prefValue;
	}
}